/*
 * burton rosenberg
 * feb 2006
 *
 * branched from trfs, strfs, added encryption layer.
 *
 * trivial file system
 *
 */

#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<stdio.h>
#include<unistd.h>
#include<stdlib.h>
#include<string.h>
#include<errno.h>

#include<fuse.h>
#include "trfs_e.h"


/*
 * definitions:
 *
 * location: byte offset in filesystem
 * cluster:  location divided by custer_size
 * offset:   the location mod cluster
 *
 * cluster_B: bit representation of cluster_size
 * dirent_Number: the location of a dirent, identifies a file or directory
 * 
 * to identify the end of files, so as to be able to update length, 
 * invariant shall be that the last cluster if marked with the
 * EOF type in the FAT, and contains file_length % cluster_size used
 * bytes. I.e., a file of length a multiple of cluster_size ends
 * with an entirely unused cluster marked as EOF
 *
 * identifying a file by dirent_number, is almost like an inode.
 * perhaps I will call it an inode
 *
 * the cluster_number of a file is the cluster number of the first
 * cluster in the file.
 *
 * directories are handled separately from files. they are a
 * sequence of dirent_s, each one typed. the first is a special
 * parent pointer which should not be deleted. the last is a
 * special NULL (or EOD, end of directory) marker, which needs to
 * be moved as the directory expands
 *
 * first entry of a directory (cluster aligned, called the cluster
 * number of the directory) is type SELF. cluster field of SELF is parent
 * pointer, unless we are root, in which case it points to self
 * length field of SELF is the byte length of the directory, counting
 * deleted entries, the SELF entry, the NULL entry, but not unused
 * space at end of last cluster.
 *
 */

struct superblock_s sb_g = { } ;

int clusterToLoc( int clusternum ) {
  return clusternum*sb_g.cluster_size ;
}

int locToCluster( int bytenum ) {
  return bytenum/sb_g.cluster_size ;
}

int setFatEntry( int entnum, int value ) {
  int er ;
  er = LSEEK_E( sb_g.backing_file, 
     sb_g.fat_base+entnum*sizeof(struct fatent_s), SEEK_SET ) ;
  if ( er==-1 ) return -1 ;
  er = WRITE_E( sb_g.backing_file, &value, sizeof(value) ) ;
  if ( er==-1 ) return -1 ;
  return 0 ;
}

int getFatEntry(int clust) {
   struct fatent_s fe ;
   LSEEK_E(sb_g.backing_file, 
       sb_g.fat_base+clust*sizeof(struct fatent_s), SEEK_SET) ;
   READ_E( sb_g.backing_file, &fe, sizeof(fe) ) ;
   return fe.cluster ;
}

int getAndMarkFreeCluster() {
    int i ;
    for ( i=0; i<sb_g.fat_n; i++ ) {
      if (getFatEntry(i)==FAT_TYPE_FREE) { 
        /* lock in immediately */
        setFatEntry(i, FAT_TYPE_EOF) ;
        return i ;
      }
    }
    return -1 ;
}

int writeSuperblock() {
  int er ;
  er = LSEEK_E( sb_g.backing_file, SUPERBLOCK_OFFSET, SEEK_SET ) ;
  if ( er==-1 ) return -1 ;
  er = WRITE_E( sb_g.backing_file, &sb_g, sizeof(sb_g) ) ;
  if ( er==-1 ) return -1 ;
  return 0 ;
}

int readSuperblock( int fdes ) {
   int er ;
   er = LSEEK_E(fdes, SUPERBLOCK_OFFSET, SEEK_SET ) ;
   if ( er==-1 ) return -1 ;
   er = READ_E( fdes, &sb_g, sizeof(sb_g) ) ;
   if ( er==-1 ) return -1 ;
   sb_g.backing_file = fdes ;
   return 0 ;
}

int getDirent( int direntLoc, struct dirent_s * de_p ) {
   LSEEK_E( sb_g.backing_file, direntLoc, SEEK_SET ) ;
   return READ_E(  sb_g.backing_file, de_p, sizeof(struct dirent_s) ) ;
}

int setDirent( int direntLoc, struct dirent_s * de_p ) {
   LSEEK_E( sb_g.backing_file, direntLoc, SEEK_SET ) ;
   return WRITE_E(  sb_g.backing_file, de_p, sizeof(struct dirent_s) ) ;
}

int createRootDir() {

  struct dirent_s de ;
  int direntLoc ;
  int er ;

  direntLoc = clusterToLoc( sb_g.root_directory ) ;

  memset(&de,0,sizeof(de)) ;
  de.type = DIR_TYPE_SELF ;
  de.length = 2 * sizeof(struct dirent_s) ;
  de.cluster = sb_g.root_directory ;
  er = setDirent( direntLoc, &de ) ;
  if ( er == -1 ) return -1 ;

  direntLoc += sizeof(struct dirent_s) ;
  memset(&de,0,sizeof(de)) ;
  de.type = DIR_TYPE_NULL ;
  er = setDirent( direntLoc, &de ) ;
  if ( er==-1 ) return -1 ;

  return 0 ;

}

int appendEmptyDirent( int loc ) {
   int clust ;
   struct dirent_s de ;
   int new_clust ;

   clust = loc / sb_g.cluster_size ;
   if ( ((loc+sizeof(struct dirent_s)) & sb_g.offset_mask ) == 0 ) { 
       new_clust = getAndMarkFreeCluster() ;
       setFatEntry( clust, new_clust ) ;
       clust = new_clust ;
       loc = clust* sb_g.cluster_size ;
   }
   else { loc += sizeof(struct dirent_s) ; }

   memset( &de, 0, sizeof(de) ) ;
   de.type = DIR_TYPE_NULL ;
   setDirent( loc, &de ) ;

}

int getFreeDirent( int clust ) {

/* scans for a free directory entry. returns a location
   for it; if necessary, extends directory with a new
   NULL ending
*/

   int next_clust ;
   int offset ;
   int loc ;
   struct dirent_s de ;
   int t ;
   int start_clust ;
 
   start_clust = clust ;
   offset = 0 ;

   t = 1 ;
   while ( t ) {
      if ( offset >= sb_g.cluster_size ) {
         next_clust = getFatEntry(clust) ;
         if ( next_clust == FAT_TYPE_EOF ) {
            /* directory ends suddenly, should fix */
            loc = -1 ;
            break ;
         }
         clust = next_clust ;
         offset = 0 ;
      }
      loc = clust*sb_g.cluster_size + offset ;
      getDirent( loc, &de ) ;
      if ( (de.type == DIR_TYPE_NULL) 
           || (de.type == DIR_TYPE_DEL) ) break ;
      offset += sizeof(struct dirent_s) ;
   }
  
   if ( de.type == DIR_TYPE_NULL ) {
      /* extend the directory by one block */
      appendEmptyDirent( loc ) ;
      /* update length */
      getDirent( clusterToLoc(start_clust), &de ) ;
      if ( de.type != DIR_TYPE_SELF )  {
         printf("getFreeDirent: %d\n", __LINE__) ;
         return loc ; /* error! */
      }
      de.length += sizeof(struct dirent_s) ;
      setDirent( clusterToLoc(start_clust), &de ) ;
   }

   return loc ;
}

int nextDirent( int dirent ) {
   dirent += sizeof( struct dirent_s ) ;
   if ( (dirent&sb_g.offset_mask) == 0 ) {
      dirent = clusterToLoc( getFatEntry( locToCluster( dirent-1 ) ) ) ;
   }
   return dirent ;
}

int findInDirByName( int dir_cluster, char * name ) {
   struct dirent_s de ;
   int loc ;
   loc = clusterToLoc( dir_cluster ) ;
   /* walk loc */
   while ( 1 ) { 
     getDirent( loc, &de ) ;
     if ( de.type==DIR_TYPE_NULL ) break ;
     if ( de.type!=DIR_TYPE_DEL ) {
        if ( strcmp( de.name, name ) == 0 ) break ;
     }
     loc = nextDirent( loc ) ;
   }
   if ( de.type==DIR_TYPE_NULL ) return 0 ;
   return loc ;
}

#define PATH_PARSE_CHARS "/"

char * getBaseName( const char * pathName ){
   char * p ;
   char * p2 ;
   p = strdup(pathName) ;
   if (p==NULL) return p ;
   p2 = strrchr(p, '/' ) ;
   if ( p2==p ) {
     *(p2+1) = '\0' ;
   } else {
     *p2 = '\0' ;
   }

   return p ;
}

/**  function to shorten a directory ...
 *   given a cluster number, walks dirents until it finds
 *   NULL
**/

int releaseFromThisCluster( int st_cluster ) {

   int clust ;
   int next_clust ;

   clust = getFatEntry( st_cluster ) ;
   if ( clust==FAT_TYPE_EOF ) return 0 ;

   setFatEntry( st_cluster, FAT_TYPE_EOF ) ;
   while ( next_clust = getFatEntry( clust ) ) {
      setFatEntry( clust, FAT_TYPE_FREE ) ;
      if ( next_clust == FAT_TYPE_EOF ) break ;
      clust = next_clust ;
   }

   return 0 ;

}


int shortenDirectory( int dir_cluster ) {
   struct dirent_s de ;
   int loc ;
   int delLoc ;
   int flag ;
   int nentries ;
   int ndeleted ;

   flag = 0 ;
   nentries = 0 ;
   ndeleted = 0 ;
   loc = clusterToLoc( dir_cluster ) ;
   while(1) {
      nentries++ ;
      ndeleted++ ;
      getDirent( loc, &de ) ;
      if ( de.type==DIR_TYPE_NULL ) break ;
      if ( de.type==DIR_TYPE_DEL && flag==0 ) {
          delLoc = loc ;
          flag = 1 ;
          ndeleted = 1 ;
      }
      if ( de.type!=DIR_TYPE_DEL ) flag = 0 ;
      loc = nextDirent( loc ) ;
   }
   if ( flag==1 ){
      getDirent( delLoc, &de ) ;
      de.type=DIR_TYPE_NULL ;
      setDirent( delLoc, &de ) ;
      /* release clusters */
      
      releaseFromThisCluster( locToCluster(delLoc) ) ;

      /* nentreis - ndeleted */
      loc = clusterToLoc( dir_cluster ) ;
      getDirent( loc , &de ) ;
      de.length = (1+nentries-ndeleted)*sizeof(struct dirent_s) ;
      setDirent( loc, &de ) ;
   }
   return 1 ;
}

int pathToDirent( char * path, int current_dir_cluster ) {
/* if path is absolute, current_dir_cluster is ignored */
/* dot and dot-dot need to be handled .. */
/* this must get much more sophisticated ... */

    int curr_dirClust ;
    int dirent ;
    char * pc ;
    struct dirent_s de ;


    curr_dirClust = sb_g.root_directory ;
    pc = strtok( path, PATH_PARSE_CHARS ) ;
    if ( pc==NULL ) {
       /* root of filesystem requested */
       return clusterToLoc( curr_dirClust ) ;
    }
    dirent = findInDirByName( curr_dirClust, pc ) ;
    if ( dirent<=0 ) goto return_0 ;

    while ( (pc = strtok(NULL, PATH_PARSE_CHARS)) != NULL ) {
       getDirent( dirent, &de ) ;
       if ( de.type != DIR_TYPE_DIR ) goto return_0 ;
       curr_dirClust = de.cluster ;
       dirent = findInDirByName( curr_dirClust, pc ) ;
       if ( dirent<=0 ) goto return_0 ;
    }
    return dirent ;

return_0:
    return 0 ;
}

int newDir( int dir_cluster, char * dir_name ) {

   int direntLoc ;
   int loc ;
   struct dirent_s de ;

   direntLoc = getFreeDirent( dir_cluster ) ;
   memset( &de, 0, sizeof(de) ) ;
   de.type = DIR_TYPE_DIR ;
   de.cluster = getAndMarkFreeCluster() ;
   strcpy( de.name, dir_name ) ;
   setDirent( direntLoc, &de ) ;

   loc = clusterToLoc(de.cluster) ;
   memset( &de,0,sizeof(de) ) ;
   de.type = DIR_TYPE_SELF ;
   de.cluster = dir_cluster ; /* parent pointer */
   de.length = 2 * sizeof(struct dirent_s) ;
   setDirent( loc, &de ) ;

   loc += sizeof(struct dirent_s) ;
   memset( &de,0,sizeof(de) ) ;
   de.type = DIR_TYPE_NULL ;
   setDirent( loc, &de ) ;

   return direntLoc ;
   
}

int newFile( int dir_cluster, char * file_name ) {
/* creates a zero length regular file in 
   directory dir_inode, with name file_name 
 
   returns a dirent_loc

 */

   int direntLoc ;
   struct dirent_s de ;

   direntLoc = getFreeDirent( dir_cluster ) ;
   memset( &de, 0, sizeof(de) ) ;
   de.type = DIR_TYPE_FILE ;
   de.cluster = getAndMarkFreeCluster() ;
   strcpy( de.name, file_name ) ;
   setDirent( direntLoc, &de ) ;

   return direntLoc ;
   
}

int seekFile( int direntLoc, int offset ) {
/*
   finds locations in files, for now, just finds the
   location of the end of file
 */

   int loc ;
   int count ;
   struct dirent_s de ;

   getDirent( direntLoc, &de ) ;
   count = offset ;
   if ( offset==-1  ) count = de.length ;
   loc = clusterToLoc(de.cluster) ;
   while ( count>=sb_g.cluster_size ) {
     count -= sb_g.cluster_size ;
     loc = clusterToLoc( getFatEntry(locToCluster(loc)) ) ;
   }
   loc += count ;
   return loc ;

}


/* this needs testing for the write over cluster logic, especially
   the internal/external write logic, i.e., when writing to a
   cluster which is not the final versus one which is
*/

/* NOT USED , see different version below. I'd have to have
   debugged this, and the other version seems to work already
*/

int writeFileAtLoc( int direntLoc, int loc, char * buf, int count ) ;

int writeFileAtLoc_( int direntLoc, int loc, char * buf, int count ) {
/* 
   write count bytes from buf into file beginning at loc. also
   given is the loc of the dirent for this file, because we
   need to reference the file length

   note: this is reentrant

   also note: this is a rough function to use, since it expects
   loc to be refenced to the file system, not to bytes-in-file.
*/
   int countAvail ;
   struct dirent_s de ;
   int next_clust ;
   int cur_clust ;

   /* identify if this is an internal write */
   cur_clust = locToCluster( loc ) ;
   while ( getFatEntry(cur_clust)!=FAT_TYPE_EOF ) {
       countAvail = sb_g.cluster_size-(loc & sb_g.offset_mask) ;
       if ( count>countAvail ) {
         /* write out to the end of current cluster */
         writeFileAtLoc( direntLoc, loc, buf, countAvail ) ;
         loc += countAvail ;
         buf += countAvail ;
         count -= countAvail ;
       }
       else {
          /* this completes the write, within current cluster */
           LSEEK_E( sb_g.backing_file, loc, SEEK_SET ) ;
           return WRITE_E (sb_g.backing_file, buf, count ) ;
       }
       /* write continues with next cluster */
       cur_clust = locToCluster( loc ) ;
   }

   /* writes which might expand file */
   getDirent( direntLoc, &de ) ;
   countAvail = sb_g.cluster_size - ( de.length & sb_g.offset_mask ) ;
   while ( count>countAvail ) {
      /* finish up current cluster, which is files last cluster */
      writeFileAtLoc( direntLoc, loc, buf, countAvail ) ;
      loc += countAvail ;
      buf += countAvail ;
      count -= countAvail ;
      countAvail = sb_g.cluster_size ;
   }
   if ( count==countAvail ) {
      /* expand file */
      next_clust = getAndMarkFreeCluster() ;
      setFatEntry( locToCluster(loc), next_clust ) ;
   }
   /* do the write */
   LSEEK_E( sb_g.backing_file, loc, SEEK_SET ) ;
   WRITE_E( sb_g.backing_file, buf, count ) ;
   de.length += count ;
   setDirent( direntLoc, &de ) ;

   return 0 ;
}


int writeFileAtLoc( int direntLoc, int loc, char * buf, int count ) {
    
    struct dirent_s de ;
    int length ;
    int fe ;
    int res = 0 ;
    int countAvail ;
    int next_clust ;
    
    while (1) {

       /* how many bytes are available? */
       getDirent( direntLoc, &de ) ;
       fe = getFatEntry( locToCluster(loc ) ) ;
       if ( fe!=FAT_TYPE_EOF ) {
         /* case: internal write in file */
         countAvail = sb_g.cluster_size - ( loc & sb_g.offset_mask ) ;
       }
       else {
         countAvail = sb_g.cluster_size - ( de.length & sb_g.offset_mask ) ;
       }
 
       if ( count<=countAvail ) {
          /* write and leave */
          LSEEK_E( sb_g.backing_file, loc, SEEK_SET ) ;
          res += WRITE_E( sb_g.backing_file, buf, count ) ;
          if ( count==countAvail && fe==FAT_TYPE_EOF ) {
             /* need to expand cluster list */
             next_clust = getAndMarkFreeCluster() ;
             setFatEntry( locToCluster(loc), next_clust ) ;
          }
          de.length += count ;
          setDirent( direntLoc, &de ) ;
          break ;
       }

       else /* count>countAvail */ {
          writeFileAtLoc_( direntLoc, loc, buf, countAvail ) ;
          buf += countAvail ;
          loc += countAvail ;
          count -= countAvail ;
          res += countAvail ;
       }
    }
    return res ;
}


int readFileAtLoc( int direntLoc, int loc, char * buf, int count ) {
    
    struct dirent_s de ;
    int length ;
    int fe ;
    int res = 0 ;
    int countAvail ;
    
    getDirent( direntLoc, &de ) ;
    
    while (1) {
       /* how many bytes are available? */
       fe = getFatEntry( locToCluster(loc ) ) ;
       if ( fe!=FAT_TYPE_EOF ) {
         /* case: internal read in file */
         countAvail = sb_g.cluster_size - ( loc & sb_g.offset_mask ) ;
       }
       else {
         countAvail = sb_g.cluster_size - ( de.length & sb_g.offset_mask ) ;
       }
 
       if ( count<=countAvail ) {
          /* read and leave */
          LSEEK_E( sb_g.backing_file, loc, SEEK_SET ) ;
          res += READ_E(  sb_g.backing_file, buf, count ) ;
          break ;
       }
       else {
          /* read available, update everything, and repeat */
          readFileAtLoc( direntLoc, loc, buf, countAvail ) ;
          buf += countAvail ;
          loc += countAvail ;
          count -= countAvail ;
          res += countAvail ;
       }
    }
    return res ;
}


void createTRFS( char * backing_file_name ) {
   
   int bf ;
   int i ;
   int fat_size ;
   int er ;

   bf = OPEN_E( backing_file_name, O_RDWR | O_TRUNC | O_CREAT, 0666 ) ;
   if ( bf==-1 ) {
      printf("error: %s %d\n", __FILE__, __LINE__ ) ;
      exit(0) ;
   }
 
   sb_g.cluster_size = CLUSTER_SIZE_D ;
   sb_g.fat_n = TOTAL_CLUSTERS_D ;
   sb_g.offset_mask = sb_g.cluster_size - 1 ;
   fat_size = (sb_g.fat_n * sizeof(struct fatent_s)
              + sb_g.cluster_size - 1) / sb_g.cluster_size ;
   sb_g.root_directory =  fat_size+1 ;
   sb_g.fat_base = sb_g.cluster_size ;
   sb_g.backing_file = bf ;

   if ( writeSuperblock() == -1 ) {
      printf("Error: %s %d\n", __FILE__, __LINE__ ) ;
      exit(0) ;
   }

   for ( i=0; i<sb_g.fat_n; i++ ) {
     setFatEntry( i, FAT_TYPE_FREE ) ;
   }
   for ( i=0; i<sb_g.root_directory; i++ ) {
     setFatEntry( i, FAT_TYPE_EOF ) ;
   }
   setFatEntry( sb_g.root_directory, FAT_TYPE_EOF ) ;
   if (createRootDir()  == -1 ) {
      printf("Error: %s %d\n", __FILE__, __LINE__ ) ;
      exit(0) ;
   }

   CLOSE_E(sb_g.backing_file) ;
}

int openTRFS( char * backing_store_name ) {

   int bf ;
   int i ;
   int fat_size ;
   int er ;

   bf = OPEN_E( backing_store_name,O_RDWR, 0666 ) ;
   if ( bf==-1 ) {
      printf("error: %s %d\n", __FILE__, __LINE__ ) ;
      exit(0) ;
   }

   if (readSuperblock(bf) == -1 ) {
      printf("error: %s %d\n", __FILE__, __LINE__ ) ;
      exit(0) ;
   }

   return 0 ;

}

void dumpFat() {
  
   int loc ;
   int indx ;
   int fatent ;
   int cntzeros ;

   cntzeros = 0 ;

   printf("\nSuperblock\n----------\ncluster size: %d\nfat entries: %d"
      "\nroot dir: %d\n\n", 
      sb_g.cluster_size, sb_g.fat_n, sb_g.root_directory ) ;

   printf("  FAT  \nindex  value\n------------\n") ;
   for ( loc = sb_g.fat_base, indx=0 ; ;loc+=sizeof(int), indx++ )    {
      LSEEK_E( sb_g.backing_file, loc, SEEK_SET ) ;
      READ_E(  sb_g.backing_file, &fatent, sizeof(int) ) ;
      printf("%4d %6d\n", indx, fatent ) ;
      if (fatent==0) cntzeros++ ;
      if ( cntzeros>10 ) break ;
   }
}


/*

 ***********
 * fuse
 ***********

 */

static int trfs_getattr(const char *path, struct stat *stbuf) {

printf("trfs_getattr: %s\n", path) ;
    int dirent ;
    struct dirent_s de ;
    char * p ;

    memset(stbuf, 0, sizeof(struct stat));
    p = strdup(path) ;
    dirent = pathToDirent( p, 0 ) ;
    free(p) ;
    if ( dirent<=0 ) return -ENOENT ;

    getDirent( dirent, &de ) ;
    switch ( de.type ) {
    case DIR_TYPE_FILE:
        stbuf->st_mode = S_IFREG | 0444;
        stbuf->st_size = de.length ;
        break ;
    case DIR_TYPE_SELF:
    case DIR_TYPE_DIR:
        stbuf->st_mode = S_IFDIR | 0755;
        getDirent( clusterToLoc( de.cluster ), &de ) ;
        /* ASSERT: de.type == DIR_TYPE_SELF */
        stbuf->st_size = de.length ;
        break ;
    case DIR_TYPE_DEL:
        /* should not happen */
    default: ;
    }
    stbuf->st_nlink = 1;

    return 0 ;
}

static int trfs_readdir(const char *path, void *buf, 
     fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi) {
printf("trfs_readdir: path = %s\n", path) ;
    char * p ;
    int dirent ;
    struct dirent_s de ;

    p = strdup(path) ;
    dirent = pathToDirent(p,0 ) ;
    free(p) ;
    if ( dirent==0 ) return -ENOENT ;
    getDirent( dirent, &de ) ;
    /* in the case of the root, dirent points to parent pointer.
       this is an argument for not distinctly typing parent different
       that dir  */

    /* this gets confusing ... but it looks like the first dirent of
       a directory is of type parent, has a cluster pointing to the
       parent, unless it is root, then it points to itself, and its
       attributes should be those of this directory. */

    if ( (de.type != DIR_TYPE_DIR) 
         && (de.type != DIR_TYPE_SELF) )  return -ENOENT ;

    /* walk directory */
    dirent = clusterToLoc( de.cluster ) ;
    getDirent( dirent, &de ) ;
    while (1) {
       if ( de.type == DIR_TYPE_SELF )  {
          filler(buf, ".", NULL, 0);
          filler(buf, "..", NULL, 0);
       }
       if ( de.type == DIR_TYPE_DIR ) {
          filler(buf, de.name, NULL, 0 ) ;
       }
       if ( de.type == DIR_TYPE_FILE ) {
          filler(buf, de.name, NULL, 0 ) ;
       }
       dirent = nextDirent( dirent ) ;
       getDirent( dirent, &de ) ;
       if ( de.type == DIR_TYPE_NULL ) break ;
    }

    return 0 ;
}

static int trfs_open(const char *path, struct fuse_file_info *fi) {
printf("trfs_open:\n") ;
    return 0 ;
}

static int trfs_read(const char *path, char *buf, 
    size_t size, off_t offset, struct fuse_file_info *fi) {

    int dirent ;
    struct dirent_s de ;
    int loc ;
    char * p ;

printf("trfs_read:\n") ;
 
    p = strdup(path) ;
    dirent = pathToDirent( p, 0 ) ;
    free(p) ;
    if ( dirent==0 ) return -1 ;
    getDirent( dirent, &de ) ;
    if ( de.type != DIR_TYPE_FILE )  return -1 ;
    loc = seekFile(dirent , offset) ;
    return readFileAtLoc( dirent, loc, buf, size ) ;

}

static int trfs_mkdir(const char *path, mode_t mode) {
    char * p ;
    char * dirname ;
    struct dirent_s de ;
    int dirent ;

printf("trfs_mkdir: %s\n", path ) ;
    p = strdup(path) ;
    dirent = pathToDirent( p, 0 ) ;
    free(p) ;
    if (dirent!=0) {
       /* directory exists */
       return -EEXIST ;
    }

    p = strdup(path) ;
    dirname = strrchr(p,'/') ;
    if ( dirname==NULL ) return -ENOENT ;
    *dirname = '\0' ;
    dirname++ ;
printf("p=|%s| dirname=|%s| LINE=%d\n",p, dirname, __LINE__ ) ;
    dirent = pathToDirent( p, 0 ) ;
printf("dirent= %d %d\n", dirent, __LINE__ ) ;
    getDirent( dirent, &de ) ;
    if ( de.type != DIR_TYPE_DIR && de.type != DIR_TYPE_SELF ) 
        return -ENOTDIR ;
    newDir(de.cluster, dirname ) ;
    free(p) ;

    return 0 ;
}

static int trfs_mknod( const char * path, mode_t mode, dev_t dev ) {
    char * p ;
    char * fn ;
    int dirent ;
    struct dirent_s de ;

    p = strdup(path) ;
    fn = strrchr(p,'/') ;
    if (fn==NULL) return -ENOENT ;
    *fn = '\0' ;
    fn++ ;
    dirent = pathToDirent( p, 0 ) ;
    if ( dirent<0 ) {
       free(p) ;
       return dirent ; /*error*/
    }
    if ( dirent==0 ) /* we are dealing w/ the root */
       dirent = clusterToLoc( sb_g.root_directory ) ;
    getDirent( dirent, &de ) ;
    newFile( de.cluster, fn ) ;
    free(p) ;
 
    return 0 ;
}


static int trfs_write( const char * path , const char * buf, size_t size, 
    off_t offset, struct fuse_file_info * ffi ) {
   
    char * p ;
    int dirent ;
    struct dirent_s de ;
    int loc ;
    
    p = strdup( path ) ;
    dirent = pathToDirent( p, 0 ) ;
    free( p ) ;
    if ( dirent<0 ) return dirent ;
    if ( dirent==0 ) return -1 ;
    getDirent( dirent, &de ) ;
    if ( de.type != DIR_TYPE_FILE ) return -EISDIR ;
    loc = seekFile(dirent , offset) ;
    /* cast buf to avoid warning concerning const qualifier */
    writeFileAtLoc( dirent, loc, (char *) buf, size ) ;
    return size ;
}

static int trfs_unlink( const char * path ) {
printf("trfs_unlink: %s\n", path ) ;

    char * p ;
    struct dirent_s de ;
    int dirent ;
    int clust, next_clust ;

    p = strdup( path ) ;
    dirent = pathToDirent( p, 0 ) ;
    free(p) ;
    if ( dirent<0 ) return dirent ;
    if ( dirent==0 ) return -1 ;

    getDirent( dirent, &de ) ;
    if ( de.type != DIR_TYPE_FILE ) return -EISDIR ;
    clust = de.cluster ;
    de.type = DIR_TYPE_DEL ;
    setDirent( dirent, &de ) ;

    /* release all clusters of file */
    while ( next_clust = getFatEntry( clust ) ) {
       setFatEntry( clust, FAT_TYPE_FREE ) ;
       if ( next_clust == FAT_TYPE_EOF ) break ;
       clust = next_clust ;
    }

    /* can attempt to shorten directory ... */
    p = getBaseName( path ) ;
    dirent = pathToDirent( p,0 ) ;
    free(p) ;
    if ( dirent<0 ) return dirent ;
    if ( dirent==0 ) return -1 ;
    getDirent(dirent,&de) ;
    shortenDirectory( de.cluster ) ;


    return 0 ;
}

int isEmptyDirectory( int clust ) {

    int dirent ;
    struct dirent_s de ;

    dirent = clusterToLoc( clust ) ;
    getDirent( dirent, &de ) ;
    while ( de.type != DIR_TYPE_NULL ) {
      if ( de.type == DIR_TYPE_FILE || de.type == DIR_TYPE_DIR ) return 0 ;
      dirent = nextDirent( dirent ) ;
      getDirent( dirent, &de ) ;
    }
    return 1 ;

}

static int trfs_rmdir( const char * path ) {

printf("trfs_rmdir: %s\n", path ) ;

    char * p  ;
    int dirent ;
    struct dirent_s de ;
    int clust, next_clust ;

    /* get dirent for this directory */
    p = strdup(path) ;
    dirent = pathToDirent( p, 0 ) ;
    free(p) ;
    if ( dirent<0 ) return dirent ;
    if ( dirent==0 ) return -1 ;
    getDirent( dirent, &de ) ;
/* get correct error code */
    if ( de.type != DIR_TYPE_DIR ) return -1 ;

    /* check that directory is empty */
    if ( !isEmptyDirectory( de.cluster ) ) {
       return -1 ;
    }

    de.type = DIR_TYPE_DEL ;
    setDirent( dirent, &de ) ;

    clust = de.cluster ;
    /* release all clusters of file */
    while ( next_clust = getFatEntry( clust ) ) {
       setFatEntry( clust, FAT_TYPE_FREE ) ;
       if ( next_clust == FAT_TYPE_EOF ) break ;
       clust = next_clust ;
    }

    /* can attempt to shorten directory ... */
    p = getBaseName( path ) ;
    dirent = pathToDirent( p,0 ) ;
    free(p) ;
    if ( dirent<0 ) return dirent ;
    if ( dirent==0 ) return -1 ;
    getDirent(dirent,&de) ;
    shortenDirectory( de.cluster ) ;

    return 0 ;
}

static struct fuse_operations trfs_oper = {
    .getattr    = trfs_getattr,
    .readdir    = trfs_readdir,
    .open       = trfs_open,
    .read       = trfs_read,
    .mkdir      = trfs_mkdir,
    .mknod	= trfs_mknod,
    .write	= trfs_write,
    .unlink	= trfs_unlink,
    .rmdir	= trfs_rmdir,
};


/*
int main(int argc, char *argv[])
{
    return fuse_main(argc, argv, &trfs_oper);
}
*/

void usage(char * fn) {
   printf("Usage: %s [[-c|-f] filename] [-r diversity] -d mountpoint\n", fn) ;
}

int main(int argc, char * argv[]) {

  int ch ;
  int create_flag ;
  int argc_fuse ;
  int diversity ;
  char * argv_fuse[3] ;
  char * filename ;
  char * progname ;


  progname = argv[0] ;
  filename = NULL ;
  create_flag = 0 ;
  argv_fuse[0] = NULL ;
  diversity = 0 ;

  while ((ch=getopt(argc,argv,"f:c:r:d:"))!=-1) {
  switch(ch) {
  case 'c':
     create_flag = 1 ;
     /* fall through */
  case 'f':
     filename = optarg ; 
     break ;
  case 'r':
     diversity = atoi(optarg) ;
     break ;
  case 'd':
     argv_fuse[0] = argv[0] ;
     argv_fuse[1] = "-d" ;
     argv_fuse[2] = optarg ;
     argc_fuse = 3 ;
     break ;
  case '?':
  default:
     usage(progname) ;
     exit(0) ;
  }}
  argc -=optind ;
  argv +=optind ;
     
  if (filename==NULL || argv_fuse[0]==NULL) {
     usage(progname) ;
     exit(0) ;
  }

  if (diversity<1) diversity = N_SPLIT_D ;
  if (diversity>4) diversity = 4 ;

  SSIO_INIT_E( diversity, CLUSTER_SIZE_D ) ;


/*
   if ( argc>1 && strcmp(argv[1],"-c")==0 ) {
     if ( argc<=2 || argv[2]==NULL ) { 
       usage(argv[0]) ;
       exit(0) ; }
     printf("Creating file system in file |%s|.\n", argv[2]) ;
     createTRFS(argv[2]) ;
     fn = argv[2] ;
     argc -= 2 ;
     argv += 2 ; }
   else {
      if ( argc<2 ) {
         usage(argv[0]) ;
         exit(0) ; }
      fn = argv[1] ;
      argc -= 1 ;
      argv += 1 ; }
*/

   if ( create_flag ) {
      createTRFS(filename) ;
   }
   openTRFS(filename) ;

   /* attach to fuse, sending fuse arguments */
   fuse_main(argc_fuse, argv_fuse, &trfs_oper);
   /* make a sample file and write into it */

/* various other tests */

/*
   {
      int direntLoc ;
      direntLoc = newFile( sb_g.root_directory, "myfile.1" ) ;
      if ( direntLoc==0 ) {
          printf("could not make myfile.1") ;
          exit(0) ;
      }
      writeFileAtLoc( direntLoc, seekFile(direntLoc,-1), "hello world", 11 ) ;
      writeFileAtLoc( direntLoc, seekFile(direntLoc,-1), "hello world", 11 ) ;
   }
*/

/*
{
   int direntLoc ;
   newDir( sb_g.root_directory, "mydir") ;
   writeFileAtLoc( direntLoc, seekFile(direntLoc,-1), "hello world", 11 ) ;
   printf("find file myfile.1: %0x\n", 
      findInDirByName( sb_g.root_directory, "myfile.1") ) ;
   printf("find file myfile.1: %0x\n", 
      findInDirByName( sb_g.root_directory, "myfile.x") ) ;
   writeFileAtLoc( direntLoc, seekFile(direntLoc,-1), "hello world", 11 ) ;
   writeFileAtLoc( direntLoc, seekFile(direntLoc, 0), "goodby world", 12 ) ;
   newFile( sb_g.root_directory, "myfile.2" ) ;
   newFile( sb_g.root_directory, "myfile.3" ) ;
   dumpFat() ;
   printf("findfile: %s %d\n", "/myfile.1", 
       pathToDirent( "/myfile.x", 0 ) ) ;
}
*/

}

